%!octave 
%
%      
%       OS : GNU/Linux 3.9.3-1-ARCH 
% COMPILER : GNU Octave, version 3.8.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: mlp_trn_gd_stoch.m,v 1.3 2015/06/04 10:44:35 mechanoid Exp mechanoid $



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом градиентного спуска, стохастический вариант
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
%
function [W msqe_hist]=mlp_trn_gd_stoch(L,C,W),
   SWE=5; % размер окна контроля за возрастанием ошибки
   MAX_ITER=900 ; % максимальное число циклов обучения
   X=L{1} ; Y=L{2} ; % учебный набор
   szX=size(X,1) ; % количество объектов в учебном наборе
   trsp=1e-1; % скорость обучения (изменения весов)
   r=round(szX*0.9); % количество учебных примеров за одну итерацию

   %C=L; % не использовать контрольный набор
   
   msqe_hist=[ 100:-1:(100-SWE) ]; % история значений ошибки обучения
   
   for iter=1:MAX_ITER,

	  W0=W; % сохраняем старое состояние
	  for i=randperm(szX)(1:r), % случайная последовательность учебных примеров
		 g=mlp_bp_grad(W,X(i,:),Y(i,:)); % градиент ошибки сети
		 p=antigrad(g); % направление изменения параметров
		 W=weight_correct(W,p,trsp); % корректируем веса
	  endfor

 	  % вычисляем ошибку сети на контрольном наборе
	  Yr=mlp_test(C{1},W) ;
	  msqe_hist(end+1)=msqe(C{2},Yr) ;

	  % условия остановки обучения
	  if stop_condition(msqe_hist(end-SWE:end),trsp,iter,MAX_ITER), break ; endif

	  %trsp=tr_speed2(trsp,msqe_hist); % скорость обучения (изменения весов)

      de=msqe_hist(end-1)-msqe_hist(end) ; % изменение ошибки
 	  if( de < 0 ), % ошибка выросла  
		 p=1.0/(1.0+exp(de)); % вероятность возврата старого состояния в случае увеличения ошибки
		 if rand() < p, 
		    printf("\n[i] ошибка выросла, возврат старого состояния с вероятностью %f\n",p);
			W=W0; 
			msqe_hist(end)=[];
		 endif
 	  endif

   endfor

   msqe_hist(1:SWE+1)=[]; % удаляем из истории ошибок начальные значения

   if iter>(MAX_ITER-1) , printf("\nдостигнут порог количества циклов обучения         \n") ; endif

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% корректируем веса
function retval=weight_correct(W,p,s),
   for j=1:length(W), 
	  W{j}= W{j} + s * p{j} ; 
   endfor
   retval=W;
endfunction



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=tr_speed(x),
   TSP=7.0 ; % параметр изменения скорости обучения
   retval= 1.0/(x*TSP) + 1.0/TSP ;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=tr_speed2(s,er),
   k=1.01; % коэффициент допустимого роста ошибки
   ri=1.01; % коэффициент увеличения скорости
   rd=0.95; % коэффициент уменьшения скорости

   if( er(end) > k*er(end-1) ),  % существенный рост ошибки,
	  s*=rd; % уменьшаем скорость обучения
   else 
	  s=min(s*ri,1.0); % увеличиваем скорость обучения
   endif

   retval=s ;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% условия остановки обучения
function retval=stop_condition(er,st,iter,max_iter),
   ERR_MIN=8e-2 ; % порог минимальной ошибки обучения
   ERR_MIN_DIFF=1e-7; % порог минимального отклонения ошибки обучения 

   printf("%i/%i : %f / %f   , speed: %f   \r",iter,max_iter,er(end),ERR_MIN, st ) ; 

   if er(end)<ERR_MIN , 
	  printf("\nдостигнут порог минимальной ошибки обучения           \n") ;
	  retval=1;
	  return;
   endif

   if monot_inc(er) ,
	  printf("\nмонотонный рост ошибки обучения           \n") ;
	  retval=1;
	  return;
   endif

   md=max(abs(mean(er)-er)) ; % максимальное отклонение ошибки от среднего значения в окне контроля
   if md < ERR_MIN_DIFF,
	  printf("\nдостигнут порог минимального отклонения ошибки обучения          \n") ;
	  retval=1;
	  return;
   endif

   retval=0;

endfunction 

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% антиградиент
function retval=antigrad(g), 
   p={}; 
   for j=1:length(g), p{end+1}=-g{j}; endfor 
   retval=p;
endfunction



